class MultibotError(Exception):
    pass


class BotNotFoundError(MultibotError):
    """
    If the bot is not found with the given name
    """
    def __init__(self, name: str):
        super().__init__(f"{name} is not found")



from typing import Union, Any
from multiprocessing import Process, Queue, set_start_method, get_start_method
from asyncio import create_task, sleep
from time import sleep as tsleep

from .process_for_bots import ProcessBot
from .process_messages import ProcessMessage


class Multibot:

    def __init__(self, limit_bots_in_tread: int = 10):
        """
        Create an instance of Multibot_asyncio class to manage few bots with asyncio.
        :param limit_bots_in_tread: Max running bot in a single asyncio loop in a thread.
        """
        if get_start_method() != "spawn":
            set_start_method("spawn")

        self.__queue_parent: Queue = Queue()
        self.__queue_children: Queue = Queue()

        self.__process: Union[Process] = Process(target=ProcessBot, args=(self.__queue_parent, self.__queue_children, limit_bots_in_tread))  # Process to run all bots

    async def start_process(self) -> "Multibot":
        """
        Start the process. It is required !
        """
        self.__process.start()
        create_task(self.__message_process_receiver())

    def __message_process_sender(self, message: dict):
        """
        Send a message to children process
        """
        print(f"Message envoyé par le parent : {message}")
        self.__queue_children.put(message)

    async def __message_process_receiver(self):
        """
        Wait message from children process (always a dict with the key "children_message")
        """
        while True:
            await sleep(0.02)
            while not self.__queue_parent.empty():
                try:
                    message = self.__queue_parent.get_nowait()
                    await self.__decode_message(message)
                except:
                    break

    async def __decode_message(self, message: Any):
        """
        Decode the current message sent by the parent process
        :param message: The message.
        """
        print(message)

    async def add_bot(self, token: str, name: str = None, autoshared: bool = False, **kwargs):
        """
        Add a bot
        """
        data = {
         'parent_message': ProcessMessage.ADD_BOT.value,
         'token': token,
         'name': name,
         'autoshared': autoshared,
         }
        data.update(kwargs)

        self.__message_process_sender(data)

    async def run_all_bots(self):
        """
        Run all bots
        """
        self.__message_process_sender({'parent_message': ProcessMessage.RUN_ALL.value})





from discord import AutoShardedBot, Bot
from random import choice
from string import ascii_letters
from asyncio import set_event_loop, new_event_loop, sleep, run, create_task
from multiprocessing import Queue

from .errors import *
from .runner import Runner
from .process_messages import ProcessMessage


class ProcessBot:

    def __init__(self, queue_parent: Queue, queue_children: Queue, limit_bots_in_tread: int = -1):
        """
        Class to manage process and thread
        """
        self.__threads: dict = {}
        self.__all_bots: dict = {}

        self.__limit_bots_in_tread: int = limit_bots_in_tread

        self.__queue_parent: Queue = queue_parent
        self.__queue_children: Queue = queue_children
        self.__loop = None
        run(self.run_process())

    async def run_process(self):
        """
        Function run with the process
        """
        await create_task(self.__message_process_receiver())  # Run the process

    async def __message_process_receiver(self):
        """
        Wait message from parent process (always a dict with the key "parent_message")
        """
        while True:
            await sleep(0.02)
            while not self.__queue_children.empty():
                try:
                    message = self.__queue_children.get_nowait()
                    await self.__decode_message(message)
                except:
                    break

    def __message_process_sender(self, message: dict):
        """
        Send a message to parent process
        """
        self.__queue_parent.put(message)

    async def __decode_message(self, message: dict):
        """
        Decode the current message sent by the parent process
        :param message: The message.
        """
        action = message['parent_message']

        if ProcessMessage.ADD_BOT.value == action:
            self.add_bot(**message)

        elif ProcessMessage.RUN_ALL.value == action:
            print("run all bots")

    def add_bot(self, token: str, *args, autoshared: bool = False, name: str = None, **kwargs) -> None:
        """
        Add a bot in the instance to manage it.
        Use this function to set the bot's intents and other subtleties.
        :param name: The name of the bot to find it. If None, a random name is given
        :param token: The token bot
        :param autoshared: Autoshare the bot in Discord
        :param args: all arguments of the Discord class Bot.
        :param kwargs: all kwargs of the Discord class Bot
        See : https://docs.pycord.dev/en/stable/api/clients.html#discord.Bot
        """
        print(f"Ajout du bot {name} avec le token {token}")
        if name is None:
            name = ''.join([choice(ascii_letters) for _ in range(20)])

        if autoshared:
            self.__all_bots[name] = {'runner': Runner(AutoShardedBot(*args, **kwargs)),
                                     'token': token,
                                     'thread': None}

        else:
            self.__all_bots[name] = {'runner': Runner(Bot(*args, **kwargs)),
                                     'token': token,
                                     'thread': None}

        print(f"Envoi de la réponse 'coucou' pour {name}")
        self.__message_process_sender({'children_message': "coucou"})


    def get_bots_names(self) -> list[str]:
        """
        Return all bots names registered in the class.
        """
        print('coucou', self.__event)
        return list(self.__all_bots.keys())


    def get_runner(self, name: str) -> "Runner":
        """
        Run the bot runner associated with this name
        :raise: BotNotFoundError
        """

        if name not in self.__all_bots.keys():
            raise BotNotFoundError(name)

        return self.__all_bots[name]['runner']



from discord import AutoShardedBot, Bot
from typing import Union

class Runner:

    def __init__(self, bot: Union[AutoShardedBot, Bot]):
        """
        Run the bot
        """
        self.bot: Union[AutoShardedBot, Bot] = bot

    def __reduce__(self):
        # Renvoyer un tuple avec les informations nécessaires pour reconstruire l'objet
        return self.__class__, (self.bot,)
